@@ -195,7 +195,11 @@ module INotify
         d = File.join(path, base)
         binary_d = d.respond_to?(:force_encoding) ? d.dup.force_encoding('BINARY') : d
         next if binary_d =~ /\/\.\.?$/ # Current or parent directory
-        watch(d, *flags, &callback) if !RECURSIVE_BLACKLIST.include?(d) && File.directory?(d)
+        next if RECURSIVE_BLACKLIST.include?(d)
+        next if flags.include?(:dont_follow) && File.symlink?(d)
+        next if !File.directory?(d)
+
+        watch(d, *flags, &callback)
       end
 
       dir.close
@@ -235,7 +239,10 @@ module INotify
     #
     # @see #run
     def process
-      read_events.each {|event| event.callback!}
+      read_events.each do |event|
+        event.callback!
+        event.flags.include?(:ignored) && event.notifier.watchers.delete(event.watcher_id)
+      end
     end
 
     # Close the notifier.
@@ -264,7 +271,7 @@ module INotify
     #
     # {#run} or {#process} are ususally preferable to calling this directly.
     def read_events
-      size = 64 * Native::Event.size
+      size = Native::Event.size + Native.fpathconf(fd, Native::Flags::PC_NAME_MAX) + 1
       tries = 1
 
       begin
@@ -298,10 +305,14 @@ module INotify
       # Use Ruby's readpartial if possible, to avoid blocking other threads.
       begin
         return to_io.readpartial(size) if self.class.supports_ruby_io?
-      rescue Errno::EBADF
+      rescue Errno::EBADF, IOError
         # If the IO has already been closed, reading from it will cause
-        # Errno::EBADF.
+        # Errno::EBADF. In JRuby it can raise IOError with invalid or
+        # closed file descriptor.
         return nil
+      rescue IOError => ex
+        return nil if ex.message =~ /stream closed/
+        raise
       end
 
       tries = 0
